@@ -17,6 +17,7 @@ import imp
 import logging
 import os
 import codecs
+import abc
 
 
 def get_path(): # XXX
@@ -24,7 +25,7 @@ def get_path(): # XXX
     return os.path.join(os.fsencode(ninix.__path__[0]), b'dll')
 
 
-class SAORI:
+class SAORI(metaclass=abc.ABCMeta):
 
     RESPONSE = {204: b'SAORI/1.0 204 No Content\r\n\r\n',
                 400: b'SAORI/1.0 400 Bad Request\r\n\r\n',
@@ -75,6 +76,7 @@ class SAORI:
         else:
             return self.RESPONSE[400]
 
+    @abc.abstractmethod
     def execute(self, args):
         return None
 
