@@ -98,6 +98,7 @@ class Sakura:
         self.script_queue = []
         self.script_mode = self.BROWSE_MODE
         self.script_post_proc = []
+        self.script_finally = []
         self.event_queue = []
         self.__current_script = ''
         self.__balloon_life = 0
@@ -123,6 +124,16 @@ class Sakura:
         self.__dll = ninix.dll.Library('shiori', saori_lib=saori_lib)
         self.__temp_mode = 0
         self.__observers = {}
+        self.__listening = {
+            'OnOtherGhostBooted': True,
+            'OnOtherGhostClosed': True,
+            'OnOtherGhostChanged': True,
+            'OnOtherSurfaceChange': False,
+            'OnOtherGhostVanished': True,
+            'OnOtherGhostTalk': False,
+            'OnOtherOverlap': True,
+            'OnOtherOffscreen': True
+            }
         self.balloon = ninix.balloon.Balloon()
         self.balloon.set_responsible(self.handle_request)
         self.surface = ninix.surface.Surface()
@@ -212,6 +223,7 @@ class Sakura:
             self.set_SSP_mode(1)
         else:
             self.set_SSP_mode(0)
+        self.last_script = None
 
     def set_SSP_mode(self, flag): # XXX
         self.__sender = 'SSP' if flag else 'ninix-aya'
@@ -312,6 +324,10 @@ class Sakura:
                                 event_type='NOTIFY')
 
     def finalize(self):
+        if self.script_finally: # XXX
+            for proc in self.script_finally:
+                proc(flag_break=False)
+            self.script_finally = []
         if not self.__temp_mode:
             self.shiori.unload()
         self.stop()
@@ -323,6 +339,9 @@ class Sakura:
     def leave_temp_mode(self):
         self.__temp_mode = 0
 
+    def is_listening(self, key):
+        return self.__listening.get(key)
+
     def on_audio_message(self, bus, message):
         if message is None: # XXX: workaround for Gst Version < 0.11
             if self.script_mode == self.WAIT_MODE:
@@ -365,7 +384,7 @@ class Sakura:
             surface_win = self.surface.get_window(side)
             balloon_win.set_transient_for(surface_win)
 
-    def enqueue_script(self, script, sender, handle,
+    def enqueue_script(self, event, script, sender, handle,
                        host, show_sstp_marker, use_translator,
                        db=None, request_handler=None):
         if not self.script_queue and \
@@ -374,7 +393,7 @@ class Sakura:
                 self.sstp_request_handler.send_sstp_break()
                 self.sstp_request_handler = None
             self.reset_script(1)
-        self.script_queue.append((script, sender, handle, host,
+        self.script_queue.append((event, script, sender, handle, host,
                                   show_sstp_marker, use_translator,
                                   db, request_handler))
 
@@ -502,13 +521,10 @@ class Sakura:
         self.notify_event('OnVanishSelecting')
         self.__vanish_dialog.show()
 
-    def vanish_by_myself(self):
+    def vanish_by_myself(self, next_ghost):
         self.vanished_count += 1
         self.ghost_time = 0
-        self.request_parent(
-            'NOTIFY', 'stop_sakura',
-            lambda a: self.request_parent('NOTIFY', 'vanish_sakura', a),
-            (self))
+        self.request_parent('NOTIFY', 'vanish_sakura', self, next_ghost)
 
     def get_ifghost(self):
         return ''.join((self.get_selfname(), ',', self.get_keroname()))
@@ -632,10 +648,7 @@ class Sakura:
         if result is None:
             result = ''
         if to and result:
-            def proc():
-                self.request_parent(
-                    'NOTIFY', 'send_message', to, self.get_selfname(), result)
-            communication = proc
+            communication = to
         else:
             communication = None
         return result, communication
@@ -645,7 +658,8 @@ class Sakura:
         return result
 
     ###   CALLBACK   ###
-    def notify_start(self, init, vanished, ghost_changed, name, path):
+    def notify_start(self, init, vanished, ghost_changed,
+                     name, prev_name, prev_shell, path, last_script):
         if self.__temp_mode:
             default = None
         else:
@@ -670,6 +684,10 @@ class Sakura:
                     return
             elif self.notify_event('OnVanished', name):
                 return
+            elif self.notify_event('OnGhostChanged', name, last_script,
+                                   prev_name, None, None, None, None,
+                                   pref_shell):
+                return
             self.notify_event('OnBoot', self.surface.name, default=default)
         elif ghost_changed:
             if self.ghost_time == 0:
@@ -677,7 +695,9 @@ class Sakura:
                                      None, None, None, None, None, None,
                                      self.surface.name):
                     return
-            elif self.notify_event('OnGhostChanged', name):
+            elif self.notify_event('OnGhostChanged', name, last_script,
+                                   prev_name, None, None, None, None,
+                                   prev_shell):
                 return
             self.notify_event('OnBoot', self.surface.name, default=default)
         else:
@@ -688,8 +708,8 @@ class Sakura:
             self.vanished_count += 1
             self.ghost_time = 0
             GLib.idle_add(
-                lambda a: self.request_parent('NOTIFY', 'vanish_sakura', a),
-                (self))
+                lambda a: self.request_parent('NOTIFY', 'vanish_sakura', *a),
+                (self, None))
         self.enqueue_event('OnVanishSelected', proc=proc)
         self.vanished = 1 ## FIXME
 
@@ -855,6 +875,13 @@ class Sakura:
             if event == 'OnMouseClick' and arglist[5] == 3:
                 self.request_parent(
                     'NOTIFY', 'open_popup_menu', self, arglist[5], arglist[3])
+            self.request_parent(
+                'NOTIFY', 'notify_other', self.key,
+                event, self.get_name(default=''),
+                self.get_selfname(default=''),
+                self.get_current_shell_name(),
+                False, communication,
+                None, False, script, arglist)
             return 0
         logging.debug('=> "{0}"'.format(script))
         if self.__temp_mode == 2:
@@ -866,9 +893,16 @@ class Sakura:
         self.start_script(script)
         self.balloon.hide_sstp_message()
         if event in self.boot_event:
-            self.script_post_proc.append(self.surface_bootup)
-        if communication is not None:
-            self.script_post_proc.append(communication)
+            self.script_finally.append(self.surface_bootup)
+        def proc(flag_break=False):
+            self.request_parent(
+                'NOTIFY', 'notify_other', self.key,
+                event, self.get_name(default=''),
+                self.get_selfname(default=''),
+                self.get_current_shell_name(),
+                flag_break, communication,
+                None, False, script, arglist)
+        self.script_finally.append(proc)
         return 1
 
     def get_prefix(self):
@@ -881,7 +915,8 @@ class Sakura:
         self.surface.toggle_bind(args)
 
     def get_menu_pixmap(self):
-        path_background, path_sidebar, path_foreground = \
+        path_background, path_sidebar, path_foreground, \
+            align_background, align_sidebar, align_foreground = \
             self.surface.get_menu_pixmap()
         top_dir = os.path.join(self.get_prefix(), b'ghost', b'master')
         if not os.path.exists(path_background):
@@ -902,7 +937,8 @@ class Sakura:
                 path_foreground = path
             else:
                 path_foreground = None
-        return path_background, path_sidebar, path_foreground
+        return path_background, path_sidebar, path_foreground, \
+            align_background, align_sidebar, align_foreground
 
     def get_menu_fontcolor(self):
         return self.surface.get_menu_fontcolor()
@@ -916,6 +952,9 @@ class Sakura:
     def get_current_shell(self):
         return self.shell_directory
 
+    def get_current_shell_name(self):
+        return self.shells[self.get_current_shell()].baseinfo[0]
+
     def get_default_shell(self):
         default = self.shell_directory or b'master'            
         if default not in self.shells:
@@ -952,7 +991,7 @@ class Sakura:
         logging.info('balloon {0} {1}'.format(name, path))
         self.notify_event('OnBalloonChange', name, path)
 
-    def surface_bootup(self):
+    def surface_bootup(self, flag_break=False):
         for side in [0, 1]:
             if not self.__boot[side]:
                 self.set_surface_default(side)
@@ -1072,7 +1111,8 @@ class Sakura:
             self.__surface_life = random.randint(20, 30)
             ##logging.debug('surface_life = {0:d}'.format(self.__surface_life))
 
-    def start(self, key, init, temp, vanished, ghost_changed, prev_name):
+    def start(self, key, init, temp, vanished, ghost_changed,
+              prev_self_name, prev_name, prev_shell, last_script):
         if self.is_running():
             if temp:
                 self.enter_temp_mode()
@@ -1080,8 +1120,9 @@ class Sakura:
                 if self.__temp_mode == 1:
                     self.__temp_mode = 2
                     self.load_shiori()
-                    self.notify_start(init, vanished, ghost_changed,
-                                      prev_name, '')
+                    self.notify_start(
+                        init, vanished, ghost_changed,
+                        prev_self_name, prev_name, prev_shell, '', last_script)
             return
         self.ghost_time = 0
         self.vanished_count = 0
@@ -1095,7 +1136,7 @@ class Sakura:
         assert self.shells and shell_key in self.shells
         surface_name, surface_dir, surface_desc, surface_alias, surface, surface_tooltips = \
             self.shells[shell_key].baseinfo
-        name = prev_name if ghost_changed else surface_name
+        name = prev_self_name if ghost_changed else surface_name
         self.set_surface(surface_desc, surface_alias, surface, surface_name,
                          surface_dir, surface_tooltips)
         balloon = None
@@ -1121,7 +1162,9 @@ class Sakura:
             self.load_shiori()
         self.restart()
         self.start_time = time.time()
-        self.notify_start(init, vanished, ghost_changed, name, surface_dir)
+        self.notify_start(
+            init, vanished, ghost_changed,
+            name, prev_name, prev_shell, surface_dir, last_script)
         GLib.timeout_add(10, self.do_idle_tasks) # 10[ms]
 
     def restart(self):
@@ -1188,6 +1231,10 @@ class Sakura:
             for proc in self.script_post_proc:
                 proc()
             self.script_post_proc = []
+        elif self.script_finally:
+            for proc in self.script_finally:
+                proc()
+            self.script_finally = []
         elif self.script_mode == self.SELECT_MODE:
             if self.passivemode:
                 pass
@@ -1213,7 +1260,7 @@ class Sakura:
         elif self.script_queue and not self.passivemode:
             if self.get_silent_time() > 0:
                 self.keep_silence(True) # extend silent time
-            script, sender, self.sstp_handle, \
+            event, script, sender, self.sstp_handle, \
                     host, show_sstp_marker, use_translator, \
                     self.sstp_entry_db, self.sstp_request_handler = \
                     self.script_queue.pop(0)
@@ -1224,6 +1271,15 @@ class Sakura:
                     self.balloon.hide_sstp_message()
                 # XXX: how about the use_translator flag?
                 self.start_script(script, self.FROM_SSTP_CLIENT)
+                def proc(flag_break=False):
+                    self.request_parent(
+                        'NOTIFY', 'notify_other', self.key,
+                        event, self.get_name(default=''),
+                        self.get_selfname(default=''),
+                        self.get_current_shell_name(),
+                        flag_break, None,
+                        (sender, host), not use_translator, script, ())
+                self.script_finally.append(proc)
         elif self.get_silent_time() > 0:
             if now - self.get_silent_time() > self.SILENT_TIME:
                 self.keep_silence(False)
@@ -1310,6 +1366,7 @@ class Sakura:
     def start_script(self, script, origin=None):
         if not script:
             return
+        self.last_script = script
         self.script_origin = origin or self.FROM_GHOST
         self.reset_script(1)
         self.__current_script = script
@@ -1656,6 +1713,12 @@ class Sakura:
                     self.balloon.open_passwordinputbox(args[2])
         elif args[0:2] == ['open', 'configurationdialog']:
             self.request_parent('NOTIFY', 'edit_preferences')
+        elif args[0:2] == ['change', 'balloon'] and argc > 2:
+            key = self.request_parent('GET', 'find_balloon_by_name', args[2])
+            if key is not None:
+                desc, balloon = self.request_parent(
+                    'GET', 'get_balloon_description', key)
+                self.select_balloon(key, desc, balloon)
         elif args[0:2] == ['change', 'shell'] and argc > 2:
             for key in self.shells:
                 shell_name = self.shells[key].baseinfo[0]
@@ -1668,12 +1731,20 @@ class Sakura:
             else:
                 self.request_parent(
                     'NOTIFY', 'select_ghost_by_name', self, args[2], 0)
+        elif args[0:2] == ['call', 'ghost'] and argc > 2:
+            key = self.request_parent('GET', 'find_ghost_by_name', args[2])
+            if key is not None:
+                self.request_parent('NOTIFY', 'start_sakura_cb', key, self)
         elif args[0:1] == ['updatebymyself']:
             if not self.busy(check_updateman=False):
                 self.__update()
         elif args[0:1] == ['vanishbymyself']:
             self.vanished = 1 ## FIXME
-            self.vanish_by_myself()
+            if argc > 1:
+                next_ghost = args[1]
+            else:
+                next_ghost = None
+            self.vanish_by_myself(next_ghost)
         elif args[1:2] == ['repaint']:
             if args[0:1] == ['lock']:
                 self.lock_repaint = 1
@@ -1740,6 +1811,20 @@ class Sakura:
                     gsettings.set_string('picture-options', opt)
                 else:
                     pass # not implemented yet
+        elif args[0:2] == ['set', 'otherghosttalk']:
+            if args[2] == 'true':
+                self.__listening['OnOtherGhostTalk'] = True
+            elif args[2] == 'false':
+                self.__listening['OnOtherGhostTalk'] = False
+            else:
+                pass ## FIXME
+        elif args[0:2] == ['set', 'othersurfacechange']:
+            if args[2] == 'true':
+                self.__listening['OnOtherSurfaceChange'] = True
+            elif args[2] == 'false':
+                self.__listening['OnOtherSurfaceChange'] = False
+            else:
+                pass ## FIXME
         elif args[0] == 'sound':
             command = args[1]
             if self.audio_player is None:
@@ -1794,6 +1879,13 @@ class Sakura:
                 pass ## FIXME
         elif args[0] == '*':
             self.balloon.append_sstp_marker(self.script_side)
+        elif args[0] == 'quicksession':
+            if args[1] == 'true':
+                self.quick_session = 1
+            elif args[1] == 'false':
+                self.quick_session = 0
+            else:
+                pass ## FIXME
         else:
             pass ## FIXME
 
@@ -1908,6 +2000,10 @@ class Sakura:
     def reset_script(self, reset_all=0):
         if reset_all:
             self.script_mode = self.BROWSE_MODE
+            if self.script_finally:
+                for proc in self.script_finally:
+                    proc(flag_break=True)
+                self.script_finally = []
             self.script_post_proc = []
             self.__current_script = ''
         self.processed_script = None
