@@ -311,7 +311,23 @@ class Surface:
                              'menu_foreground.png')
         name = name.replace('\\', '/')
         path_foreground = os.path.join(top_dir, os.fsencode(name))
-        return path_background, path_sidebar, path_foreground
+        align_background = self.desc.get('menu.background.alignment',
+                                         'lefttop')
+        if align_background not in ['lefttop', 'righttop', 'centertop']:
+            align_background = 'lefttop'
+        align_background = align_background[:-3].encode('ascii') # XXX
+        align_sidebar = self.desc.get('menu.sidebar.alignment',
+                                      'top')
+        if align_sidebar not in ['top', 'bottom']:
+            align_sidebar = 'top'
+        align_sidebar = align_sidebar.encode('ascii') # XXX
+        align_foreground = self.desc.get('menu.foreground.alignment',
+                                         'lefttop')
+        if align_foreground not in ['lefttop', 'righttop', 'centertop']:
+            align_foreground = 'lefttop'
+        align_foreground = align_foreground[:-3].encode('ascii') # XXX
+        return path_background, path_sidebar, path_foreground, \
+            align_background, align_sidebar, align_foreground
 
     def get_menu_fontcolor(self):
         fontcolor_r = self.desc.get_with_type('menu.background.font.color.r', int, 0)
@@ -707,10 +723,12 @@ class SurfaceWindow:
         self.window.get_window().set_background_pattern(None) # XXX
 
     def display_changed(self, screen):
-        left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
-        self.request_parent(
-            'NOTIFY', 'notify_event', 'OnDisplayChange',
-            Gdk.Visual.get_best_depth(), scrn_w, scrn_h)
+        if self.side == 0:
+            self.request_parent('NOTIFY', 'reset_position') # XXX
+            left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
+            self.request_parent(
+                'NOTIFY', 'notify_event', 'OnDisplayChange',
+                Gdk.Visual.get_best_depth(), scrn_w, scrn_h)
 
     def set_responsible(self, request_method):
         self.request_parent = request_method
@@ -775,6 +793,7 @@ class SurfaceWindow:
         self.set_surface(self.default_id)
 
     def set_surface(self, surface_id):
+        prev_id = self.surface_id
         if self.alias is not None and surface_id in self.alias:
             aliases = self.alias.get(surface_id)
             if aliases:
@@ -813,12 +832,16 @@ class SurfaceWindow:
             self.set_position(x, y)
         if self.side < 2:
             self.request_parent('NOTIFY', 'notify_observer', 'set surface')
+        w, h = self.get_surface_size(self.surface_id)
+        new_x, new_y = self.get_position()
         self.request_parent(
             'NOTIFY', 'notify_event',
             'OnSurfaceChange',
             self.request_parent('GET', 'get_surface_id', 0),
             self.request_parent('GET', 'get_surface_id', 1),
-            "{},{},{},{}".format(self.side, self.surface_id, w, h))
+            "{},{},{},{}".format(self.side, self.surface_id, w, h),
+            "{}".format(prev_id),
+            "{},{},{},{}".format(new_x, new_y, new_x + w, new_y + h))
 
     def iter_mayuna(self, surface_width, surface_height, mayuna, done):
         for surface_id, interval, method, args in mayuna.patterns:
@@ -917,6 +940,8 @@ class SurfaceWindow:
     def update_frame_buffer(self):
         new_surface = self.create_image_surface(self.seriko.base_id)
         assert new_surface is not None
+        # update collision areas
+        self.collisions = self.region[self.seriko.base_id]
         # draw overlays
         for surface_id, x, y in self.seriko.iter_overlays():
             try:
@@ -1040,8 +1065,7 @@ class SurfaceWindow:
     def set_position(self, x, y):
         self.position = (x, y)
         new_x, new_y = self.get_position()
-        if self.__shown:
-            self.window.resize_move(new_x, new_y)
+        self.window.resize_move(new_x, new_y)
         left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
         direction = 0 if x > left + scrn_w // 2 else 1
         self.direction = direction
