@@ -243,7 +243,7 @@ class SSTPControler:
         self.set_sstp_flag(sender)
         self.request_parent(
             'NOTIFY', 'enqueue_script',
-            script, sender, handle, address,
+            event, script, sender, handle, address,
             show_sstp_marker, use_translator, entry_db,
             request_handler, temp_mode=True)
 
@@ -574,8 +574,8 @@ class Application:
             'get_preference': self.prefs.get,
             'get_otherghostname': self.communicate.get_otherghostname,
             'rebuild_ghostdb': self.communicate.rebuild_ghostdb,
+            'notify_other': self.communicate.notify_other,
             'reset_sstp_flag': self.sstp_controler.reset_sstp_flag,
-            'send_message': self.communicate.send_message,
             'get_sstp_port': self.sstp_controler.get_sstp_port,
             }
         handler = handlers.get(event,
@@ -633,13 +633,13 @@ class Application:
     def enqueue_event(self, event, *arglist, **argdict):
         self.current_sakura_instance.enqueue_event(event, *arglist, **argdict)
 
-    def enqueue_script(self, script, sender, handle,
+    def enqueue_script(self, event, script, sender, handle,
                        host, show_sstp_marker, use_translator,
                        db=None, request_handler=None, temp_mode=False):
         sakura = self.current_sakura_instance
         if temp_mode:
             sakura.enter_temp_mode()
-        sakura.enqueue_script(script, sender, handle,
+        sakura.enqueue_script(event, script, sender, handle,
                               host, show_sstp_marker, use_translator,
                               db, request_handler)
 
@@ -674,11 +674,12 @@ class Application:
     def get_current_balloon_directory(self): ## FIXME
         return self.__menu_owner.get_current_balloon_directory()
 
-    def start_sakura_cb(self, key): ## FIXME
+    def start_sakura_cb(self, key, caller=None):
         sakura_name = self.ghosts[key].instance.get_selfname(default='')
         name = self.ghosts[key].instance.get_name(default='')
-        self.__menu_owner.notify_event(
-            'OnGhostCalling', sakura_name, 'manual', name, key)
+        if caller is None:
+            caller = self.__menu_owner
+        caller.notify_event('OnGhostCalling', sakura_name, 'manual', name, key)
         self.start_sakura(key, init=1) # XXX
 
     def select_sakura(self, key):
@@ -704,9 +705,12 @@ class Application:
 
     def open_popup_menu(self, sakura, button, side):
         self.__menu_owner = sakura
-        path_background, path_sidebar, path_foreground = \
+        path_background, path_sidebar, path_foreground, \
+            align_background, align_sidebar, align_foreground = \
             self.__menu_owner.get_menu_pixmap()
-        self.__menu.set_pixmap(path_background, path_sidebar, path_foreground)
+        self.__menu.set_pixmap(
+            path_background, path_sidebar, path_foreground,
+            align_background, align_sidebar, align_foreground)
         background, foreground = self.__menu_owner.get_menu_fontcolor()
         self.__menu.set_fontcolor(background, foreground)
         mayuna_menu = self.__menu_owner.get_mayuna_menu()
@@ -906,7 +910,7 @@ class Application:
         sakura = self.ghosts[key].instance
         if not sakura.is_running():
             self.start_sakura(key, init=1)
-        sakura.enqueue_script('\![updatebymyself]\e', sender,
+        sakura.enqueue_script(None, '\![updatebymyself]\e', sender,
                               None, None, 0, 0, None)
 
     def select_current_sakura(self, ifghost=None):
@@ -977,11 +981,11 @@ class Application:
             key = random.choice(keys)
         self.change_sakura(sakura, key, 'automatic', event, vanished)
 
-    def select_ghost_by_name(self, sakura, name, event=1):
+    def select_ghost_by_name(self, sakura, name, event=1, vanished=0):
         key = self.find_ghost_by_name(name)
         if key is None:
             return
-        self.change_sakura(sakura, key, 'automatic', event)
+        self.change_sakura(sakura, key, 'automatic', event, vanished)
 
     def change_sakura(self, sakura, key, method, event=1, vanished=0):
         if sakura.key == key: # XXX: needs reloading?
@@ -1022,11 +1026,18 @@ class Application:
             else:
                 ghost_changed = 1
         if ghost_changed:
-            name = self.ghosts[prev].instance.get_selfname()
+            self_name = self.ghosts[prev].instance.get_selfname()
+            name = self.ghosts[prev].instance.get_name()
+            shell = self.ghosts[prev].instance.get_current_shell_name()
+            last_script = self.ghosts[prev].instance.last_script
         else:
+            self_name = None
             name = None
+            shell = None
+            last_script = None
         sakura.notify_preference_changed()
-        sakura.start(key, init, temp, vanished, ghost_changed, name)
+        sakura.start(key, init, temp, vanished, ghost_changed,
+                     self_name, name, shell, last_script)
         self.set_menu_sensitive(key, False)
 
     def notify_preference_changed(self):
@@ -1108,7 +1119,7 @@ class Application:
                 del self.balloons[balloon_dir]
         self.balloon_menu = self.create_balloon_menu()
 
-    def vanish_sakura(self, sakura):
+    def vanish_sakura(self, sakura, next_ghost):
         # remove ghost
         prefix = sakura.get_prefix()
         for filename in os.listdir(prefix):
@@ -1125,7 +1136,10 @@ class Application:
                 except:
                     logging.error(
                         '*** REMOVE FAILED *** : {0}'.format(filename))
-        self.select_ghost(sakura, 0, vanished=1)
+        if next_ghost is not None:
+            self.select_ghost_by_name(sakura, next_ghost, vanished=1)
+        else:
+            self.select_ghost(sakura, 0, vanished=1)
         del self.ghosts[sakura.key]
 
     def select_plugin(self, item):
